package cs307.first;

import threeohseven.BitVector;
import threeohseven.ImplementationError;

/**
 * The abstract implementation of a bit vector, requries subclasses to define how the bits are stored
 * and do methods to set, unset, negate and copy at minimum. Contains helper methods for subclasses to check the bounds of any operation.
 * @author -
 */
public abstract class BitVector_23_base implements BitVector {

    private final Builder builder;
    protected boolean checked;
    protected final int min;
    protected final int max;

    protected BitVector_23_base(Builder builder, boolean checked, int min, int max) {
        this.builder = builder;
        this.checked = checked;
        this.min = min;
        this.max = max;
    }

    /**
     * Tests the BitVector to make sure it's active range is compatible with this active range.
     * @param bs The BitVector to check it's active range
     * @throws ImplementationError if this bit vector is checked and the parameter is not in this range.
     */
    protected void checkBounds(BitVector bs) {
        if (checked) {
            BitVector.Iterator iter = bs.iterator();
            int bsMin = iter.position(); //Iterator starts out at their min position.
            if (bsMin < min || bsMin > max) {
                throw new ImplementationError("Other BitVector's min is: " + bsMin + " which is smaller than this ones");
            }
        }
    }

    /**
     * If this bit vector is checked, it makes sure that the index is valid. If it is not
     * It will throw an ImplementationError
     * @param index the index to check
     * @throws ImplementationError if the index is out of the active range.
     */
    protected void checkBounds(int index) {
        if (checked && (index < min || index > max)) {
            throw new ImplementationError("Out of bounds: " + index);
        }
    }

    /**
     * Returns an iterator over this <code>BitVector</code>.
     **/
    public Builder builder() {
        return builder;
    }

    /**
     * Returns the builder that created this <code>BitVector</code>.
     **/
    public BitVector.Iterator iterator() {
        return new Iterator(this, min, max);
    }

    /**
     * Logical <code>and</code> of the two vectors. Receiver is updated in
     * place.  The operation is only valid if <code>index</code> is in the
     * active range of the <code>BitVector</code>.
     * @param bs the other <code>BitVector</code>.
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public void and(BitVector bs) {
        checkBounds(bs);
        boolean oldCheck = checked; //We have already checked our bounds, so don't check the bounds in unset(pos)
        checked = false;
        BitVector.Iterator iter = bs.iterator();
        while (iter.position() < min) { //Move the iterator to the active range
            iter.getNext();
        }
        while (iter.hasNext() && iter.position() <= max) {
            int pos = iter.skipToUnset();//T AND F = F, T AND T = T, so true doesn't affect our bits.
            if (iter.hasNext() && pos <= max) {
                iter.getNext();
                unset(pos); //F AND F = F, F AND T = F, so the value has to be false
            }
        }
        checked = oldCheck; //Set bounds checking back to the original
    }

    /**
     * Logical <code>or</code> of the two vectors. Receiver is updated in
     * place.  The operation is only valid if the two
     * <code>BitVector</code> objects are compatible.
     * @param bs the other <code>BitVector</code>.
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public void or(BitVector bs) {
        checkBounds(bs);
        boolean oldCheck = checked;//We have already checked our bounds, so don't check the bounds in set(pos)
        checked = false;
        BitVector.Iterator iter = bs.iterator();
        while (iter.position() < min) { //Move the iterator into the active range
            iter.getNext();
        }

        while (iter.hasNext() && iter.position() <= max) {
            int pos = iter.skipToSet();//F OR F = F, F OR T = T, so false doesn't affect our bits.
            if (iter.hasNext() && pos <= max) {
                iter.getNext();
                set(pos); //T OR F = F, T OR T = T, so the value has to be true
            }
        }
        checked = oldCheck;//Set bounds checking back to the original
    }

    /**
     * Logical <code>xor</code> of the two vectors. Receiver is updated in
     * place.  The operation is only valid if the two
     * <code>BitVector</code> objects are compatible.
     * @param bs the other <code>BitVector</code>.
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public void xor(BitVector bs) {
        checkBounds(bs);
        boolean oldCheck = checked; //We just checked the bounds, so dont' check them again in negate();
        checked = false;
        BitVector.Iterator iter = bs.iterator();
        while (iter.position() < min) { //Move the iterator into the active range
            iter.getNext();
        }
        while (iter.hasNext() && iter.position() <= max) {
            int pos = iter.skipToSet();//T XOR F = T, F XOR F = F, so false doesn't affect our bits.
            if (iter.hasNext() && pos <= max) {
                iter.getNext();
                negate(pos); //T XOR F = T, T XOR T = F, so the value has to be the opposite
            }
        }
        checked = oldCheck; //Set bounds checking back to the original
    }

    @Override
    public boolean equals(Object o) {
        if (o == null) {
            return false;
        }

        if (o instanceof BitVector) {
            final BitVector other = (BitVector) o;
            for (int i = min; i < max; i++) {
                try {
                    boolean result = other.test(i) == test(i); //Check each indiviual value, if they are true
                    if (!result) {
                        return false;
                    }
                } catch (ImplementationError e) {
                    return false; //Happens if other bitVector does not have the same min and max.
                }

            }
            return true;
        }

        return false; //Not instanceof BitVector
    }

    /**
     * Return a representation of this <code>BitVector</code> as a
     * string. The format is implementation specific.
     **/
    @Override
    public String toString() {
        BitVector.Iterator iter = iterator();
        StringBuilder stringBuilder = new StringBuilder();
        while (iter.hasNext()) {
            stringBuilder.append(iter.getNext() ? 1 : 0);
        }
        return stringBuilder.toString();
    }

    /**
     * An implementation of BiVector.Iterator that does not depend on any single BitVector
     * Not particularly efficient, as this Iterator does not know about the internals of the BitVector
     * it is iterating over, but is useful for debuging a BitVector.
     * (This class is not used in any of the given BitVectors, but is required to build as BitVector_23_base uses it for a default iterator)
     * @author -
     */
    private class Iterator implements BitVector.Iterator {

        private final int min;
        private final int max;
        private int index;
        private final BitVector bitVector;

        public Iterator(BitVector bv, int min, int max) {
            this.min = min;
            this.max = max;
            index = min;
            bitVector = bv;
        }

        /**
         * Returns the value for the current cursor, and advances the
         * cursor by one. When the cursor overflows the active range
         * (i.e. <code>hasNext()</code> is false) arbitrary values may be
         * returned. Implementations are not required to throw an exception.
         * @return the value of the next index in range.
         * @throws NoSuchElementError if hasNext() == false and vector is strict.
         **/
        public boolean getNext() {
            return bitVector.test(index++);
        }

        /**
         * Are there more values in the active range?
         * @return false if the end of the range has been reached.
         **/
        public boolean hasNext() {
            return index <= max;
        }

        /**
         * The value of the cursor. This value may be larger than the
         * vector's upper bound. In which case <code>hasNext()</code> will
         * return false.
         * @return current position.
         **/
        public int position() {
            return index;
        }

        /**
         * Skips to the next set bit such that <code>getNext()</code> will
         * return true, if none is found the cursor is moved to the end of
         * the vector and <code>hasNext()</code> will return false from
         * thereon.
         * @return position of the first set bit, if there is none an
         * invalid value will be returned (ie a value larger than the upper
         * bound of the array).
         **/
        @SuppressWarnings("empty-statement")
        public int skipToSet() {
            boolean val = false;
            while (hasNext() && (val = getNext()) == false);
            if (val == false) {
                return index;
            } else {
                return --index;
            }

        }

        /**
         * Skips to the next unset bit such that <code>getNext()</code>
         * will return false, if none is found the cursor is moved to the
         * end of the vector and <code>hasNext()</code> will return false
         * from thereon.
         * @return position of the first unset bit, if there is none an
         * invalid value will be returned (ie a value larger than the upper
         * bound of the array).
         **/
        @SuppressWarnings("empty-statement")
        public int skipToUnset() {

            boolean val = true;
            while (hasNext() && (val = getNext()) == true);
            if (val == true) {
                return index;
            } else {
                return --index;
            }
        }
    }
}
