package cs307.first;

import threeohseven.BitVector;

/**
 * The Idea of this class is that it will be sparsely populated, so we should
 * only initialize those buckets which have had a value set. As we want this to
 * be space-efficient, we choose to use our most efficient representation of a
 * BitVector we have, so we use BitVector_23_space_and_speed for each bucket.
 *
 * It is basically a hash table.
 * @author -
 */
public class BitVector_23_space extends BitVector_23_base {

    private final BitVector[] bitVector; //Basically an int[][], or boolean[][] i fyou want to look at it abstractly
    private static final int numBitsPerBucket = 1024; //32*32 = 2^10
    private final int numBitsInLastBucket;

    protected BitVector_23_space(Builder builder, boolean checked, int min, int max) {
        super(builder, checked, min, max);
        int difference = max - min;
        int numBuckets = (difference >> 10) + 1; //Difference/1024 + 1
        numBitsInLastBucket = getIndexInBucket(difference);//Get the last index basically
        this.bitVector = new BitVector[numBuckets];
    }

    /**
     *
     * @param index The index of the bit we are trying to get.
     * @return The last 9 bits of index, max can return is 1023 which is the largest index each bucket can contain.
     */
    private int getIndexInBucket(int index) {

        return index & 0x3ff;
    }

    /**
     *
     * @param index The index of the bit we are trying to get.
     * @return index divided by the number of bits in a bucket
     */
    private int getBucket(int index) {
        return index >> 10;
    }

    /**
     *  Get a new BitVector to put in the array of BitVectors. Each bitVector starts at zero and goes to 1023 (which is numBitsPerBucket-1)
     *  We set checked to false, because we already check the bounds in this class.
     * @return A new unchecked BitVector with a range from zero to numBitsPerBucket-1
     */
    private BitVector newBucket(int index) {
        if (index == bitVector.length) {
            return new BitVector_23_space_and_speed(null, false, 0, numBitsInLastBucket);
        } else {
            return new BitVector_23_space_and_speed(null, false, 0, numBitsPerBucket - 1);
        }
    }

    /**
     * Returns a logically disjoint copy of the object.  The operation is
     * only valid if the two <code>BitVector</code> objects are compatible.
     * @param bs the other <code>BitVector</code>.
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code>
     **/
    public BitVector copy() {
        BitVector_23_space bv = new BitVector_23_space(builder(), checked, min, max);
        for (int i = 0; i < bitVector.length; i++) {
            bv.bitVector[i] = bitVector[i] == null ? null : bitVector[i].copy();
        }
        return bv;
    }

    /**
     * Returns an iterator over this <code>BitVector</code>.
     **/
    @Override
    public BitVector.Iterator iterator() {
        return new Iterator();
    }

    /**
     * Negate the bit at <code>index</code>.  The operation is only valid
     * if <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean negate(int index) {
        checkBounds(index);
        index -= min; //"Zero" base the index.
        int bucketIndex = getBucket(index);
        BitVector bv = bitVector[bucketIndex];
        if (bv == null) {
            bv = bitVector[bucketIndex] = newBucket(bucketIndex); //If there is no bucket there, initialize one so we can flip the value.
        }
        return bv.negate(getIndexInBucket(index));
    }

    /**
     * Set the bit at <code>index</code>.  The operation is only valid if
     * <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean set(int index) {
        checkBounds(index);
        index -= min; //"Zero" base the index.
        int bucketIndex = getBucket(index);
        BitVector bv = bitVector[bucketIndex];
        if (bv == null) {
            bv = bitVector[bucketIndex] = newBucket(bucketIndex);
        }
        return bv.set(getIndexInBucket(index)); //Pass it onto the underlying bitVector
    }

    /**
     * Test the value at <code>index</code>.  The operation is only valid
     * if <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return value at <code>index</code>
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean test(int index) {
        checkBounds(index);
        index -= min; //"Zero" base the index.
        int bucketIndex = getBucket(index);
        BitVector bv = bitVector[bucketIndex];
        if (bv == null) {
            return false; //A null bucket means there is nothing in there, so it has to be false.
        }
        return bv.test(getIndexInBucket(index));
    }

    /**
     * Unset the bit at <code>index</code>.  The operation is only valid if
     * <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean unset(int index) {
        checkBounds(index);
        index -= min; //"Zero" base the index.
        int bucketIndex = getBucket(index);
        BitVector bv = bitVector[bucketIndex];
        if (bv == null) {
            return false; //This means that this bucket's values are all false, so unseting it makes it false still, so don't do anything.
        }
        return bv.unset(getIndexInBucket(index));
    }

    private class Iterator implements BitVector.Iterator {

        private int curIndex = 0; //The current bucket
        private int curBitIndex = 0; //The current index inside the bucket.

        /**
         * Returns the value for the current cursor, and advances the
         * cursor by one. When the cursor overflows the active range
         * (i.e. <code>hasNext()</code> is false) arbitrary values may be
         * returned. Implementations are not required to throw an exception.
         * @return the value of the next index in range.
         * @throws NoSuchElementError if hasNext() == false and vector is strict.
         **/
        public boolean getNext() {
            BitVector bv;
            try {
                bv = bitVector[curIndex];
            } catch (ArrayIndexOutOfBoundsException e) {
                if (checked) {
                    throw new NoSuchElementError(e.getMessage());
                } else {
                    throw e;
                }
            }
            if (bv == null) { //If there is no bucket, then the index is false.
                incrementIndex();
                return false;
            } else {
                boolean ret = bv.test(curBitIndex);
                incrementIndex();
                return ret;
            }
        }

        private void incrementIndex() {
            if (curBitIndex == numBitsPerBucket - 1) { //If its the last bit in the bucket, goto the next bucket.
                curBitIndex = 0;
                curIndex++;
            } else { //Else, increment the bitIndex
                curBitIndex++;
            }
        }

        /**
         * Are there more values in the active range?
         * @return false if the end of the range has been reached.
         **/
        public boolean hasNext() {
            return position() <= max;
        }

        /**
         * The value of the cursor. This value may be larger than the
         * vector's upper bound. In which case <code>hasNext()</code> will
         * return false.
         * @return current position.
         **/
        public int position() {
            //We multiply curIndex by 1024 because there are 1024 bits in each bucket.
            //Then we add the currentBitIndex, and then sense our internal indexes are based on zero,
            //Add min to base the index off of min.
            return (curIndex << 10) + curBitIndex + min;
        }

        /**
         * Skips to the next set bit such that <code>getNext()</code> will
         * return true, if none is found the cursor is moved to the end of
         * the vector and <code>hasNext()</code> will return false from
         * thereon.
         * @return position of the first set bit, if there is none an
         * invalid value will be returned (ie a value larger than the upper
         * bound of the array).
         **/
        public int skipToSet() {
            while (curIndex < bitVector.length) {
                BitVector curBv = bitVector[curIndex];
                if (curBv == null) {
                    curIndex++;
                    curBitIndex = 0;
                    continue;
                } else {
                    if (curBv.test(curBitIndex)) {
                        break;
                    }
                    incrementIndex();
                }
            }
            return position();
        }

        /**
         * Skips to the next unset bit such that <code>getNext()</code>
         * will return false, if none is found the cursor is moved to the
         * end of the vector and <code>hasNext()</code> will return false
         * from thereon.
         * @return position of the first unset bit, if there is none an
         * invalid value will be returned (ie a value larger than the upper
         * bound of the array).
         **/
        public int skipToUnset() {
            BitVector curBv = bitVector[curIndex];
            if (curBv == null) { //If its already false, then we are at an unset position.
                return position();
            }
            while (curIndex < bitVector.length) {
                curBv = bitVector[curIndex];
                if (curBv == null) {
                    curBitIndex = 0;
                    break;
                } else {
                    if (curBv.test(curBitIndex)) {
                        incrementIndex();
                    } else {
                        break;
                    }
                }
            }
            return position();
        }
    }
}