
package cs307.first;

import threeohseven.BitVector;
import threeohseven.ImplementationError;

/**
 * Holds bits inside of an array of integers for efficiency and speed
 * Not as fast as having an array of booleans, but 8 times more efficient (each boolean is 1 byte)
 * while still being very fast.
 * @author -
 */
public class BitVector_23_space_and_speed extends BitVector_23_base {

    //Holds (max-min)/32 + 1 integers which hold 32 bits each.
    private int[] bitVector = null;

    public BitVector_23_space_and_speed(Builder builder, boolean checked, int min, int max) {
        super(builder, checked, min, max);
        int difference = max - min;
        // Difference/32 + 1
        int arrayLen = (difference >> 5) + 1;
        bitVector = new int[arrayLen];
    }

    private int getIndex(int index) {
        return index >> 5; //The index divided by 32
    }

    /**
     * It is assumed that index has already been zero based (aka, all indexes start "min" based, to make them "zero" based, subtract min from index
     *@return The last 5 bits of index, max this can return is 31 (which is the amount of bits in an int, zero based).
     */
    private int getBit(int index) {
        return index & 0x1F;
    }

    /**
     *
     * @param bitContainer An int of bits
     * @param bitIndex The index of the bit to return
     * @return The bit inside of bitContainer at bitIndex (0 = false, 1 = true)
     */
    private boolean getValue(int bitContainer, int bitIndex) {
        int returnValue = bitContainer & (1 << bitIndex);
        return returnValue != 0;
    }

    /**
     * Returns a logically disjoint copy of the object.  The operation is
     * only valid if the two <code>BitVector</code> objects are compatible.
     * @param bs the other <code>BitVector</code>.
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code>
     **/
    public BitVector copy() {
        BitVector_23_space_and_speed bv = new BitVector_23_space_and_speed(builder(), checked, min, max);
        System.arraycopy(bitVector, 0, bv.bitVector, 0, bitVector.length);
        return bv;
    }

    /**
     * Returns an iterator over this <code>BitVector</code>.
     **/
    @Override
    public BitVector.Iterator iterator() {
        return new Iterator();
    }

    /**
     * Negate the bit at <code>index</code>.  The operation is only valid
     * if <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean negate(int index) {
        checkBounds(index);
        index -= min;
        int indexInsideBucket = getIndex(index);
        int bitIndex = getBit(index);
        boolean retValue = getValue(bitVector[indexInsideBucket], bitIndex);
        /* So lets say we have a bit container which looks like 01000101
         * and the bit index we want to negate is 3 (zero based)
         * So 00000001 shifted left by 3 is 00001000
         * Now, 01000101 XOR 00001000 is equal to 01001101.
         * So we efficiently flipped the bit.
         */
        bitVector[indexInsideBucket] ^= (1 << bitIndex);
        return retValue;
    }

    /**
     * Set the bit at <code>index</code>.  The operation is only valid if
     * <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean set(int index) {
        checkBounds(index);
        index -= min;
        int indexInsideBucket = getIndex(index);
        int bitIndex = getBit(index);
        boolean retValue = getValue(bitVector[indexInsideBucket], bitIndex);

        /* So lets say we have a bit container which looks like 01000101
         * and the bit index we want to set is 3 (zero based)
         * So 00000001 shifted left by 3 is 00001000
         * Now, 01000101 OR 00001000 is equal to 01001101.
         * So we the bit is now set.
         */
        bitVector[indexInsideBucket] |= 1 << bitIndex;
        return retValue;
    }

    /**
     * Test the value at <code>index</code>.  The operation is only valid
     * if <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return value at <code>index</code>
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean test(int index) {
        checkBounds(index);
        index -= min;
        int indexInsideBucket = getIndex(index);
        int bitIndex = getBit(index);
        return getValue(bitVector[indexInsideBucket], bitIndex);
    }

    /**
     * Unset the bit at <code>index</code>.  The operation is only valid if
     * <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean unset(int index) {
        checkBounds(index);
        index -= min;
        int indexInsideBucket = getIndex(index);
        int bitIndex = getBit(index);
        boolean retValue = getValue(bitVector[indexInsideBucket], bitIndex);
        /* So lets say we have a bit container which looks like 01000101
         * and the bit index we want to negate is 2 (zero based)
         * So 00000001 shifted left by 2 is 00000100
         * The complement of that is 11111011
         * Now, 01000101 AND 11111011 is equal to 01000001.
         * So we unset the bit.
         */
        bitVector[indexInsideBucket] &= ~(1 << bitIndex); //The complement of 1 shifted the number of bits
        return retValue;
    }

    private class Iterator implements BitVector.Iterator {

        private int index = 0;
        private int bitIndex = 0;

        /**
         * Returns the value for the current cursor, and advances the
         * cursor by one. When the cursor overflows the active range
         * (i.e. <code>hasNext()</code> is false) arbitrary values may be
         * returned. Implementations are not required to throw an exception.
         * @return the value of the next index in range.
         * @throws NoSuchElementError if hasNext() == false and vector is strict.
         **/
        public boolean getNext() {
            try {
                boolean value = getValue(bitVector[index], bitIndex);
                incrementIndex();
                return value;
            } catch (ArrayIndexOutOfBoundsException e) {
                if (checked) {
                    throw new NoSuchElementError(e.getMessage());
                } else {
                    throw e;
                }
            }
        }

        private void incrementIndex() {
            if (bitIndex == 31) { //If its the last bit in the bucket, then goto the next one (31 bits in a byte starting from zero)
                bitIndex = 0;
                index++;
            } else {
                bitIndex++;
            }
        }

        /**
         * Are there more values in the active range?
         * @return false if the end of the range has been reached.
         **/
        public boolean hasNext() {
            return max >= position();
        }

        /**
         * The value of the cursor. This value may be larger than the
         * vector's upper bound. In which case <code>hasNext()</code> will
         * return false.
         * @return current position.
         **/
        public int position() {
            return (index << 5) + bitIndex + min; // Index * 32 + bitIndex + min
        }

        /**
         * Skips to the next set bit such that <code>getNext()</code> will
         * return true, if none is found the cursor is moved to the end of
         * the vector and <code>hasNext()</code> will return false from
         * thereon.
         * @return position of the first set bit, if there is none an
         * invalid value will be returned (ie a value larger than the upper
         * bound of the array).
         **/
        public int skipToSet() {
            while (index < bitVector.length && bitVector[index] == 0) {
                index++;
                bitIndex = 0;
            }
            while (index < bitVector.length && getValue(bitVector[index], bitIndex) == false) {
                incrementIndex();
            }
            return position();
        }

        /**
         * Skips to the next unset bit such that <code>getNext()</code>
         * will return false, if none is found the cursor is moved to the
         * end of the vector and <code>hasNext()</code> will return false
         * from thereon.
         * @return position of the first unset bit, if there is none an
         * invalid value will be returned (ie a value larger than the upper
         * bound of the array).
         **/
        public int skipToUnset() {
            while (index < bitVector.length && bitVector[index] == -1) { // -1 = all 1's because of signed value.
                index++;
                bitIndex = 0;
            }
            while (index < bitVector.length && getValue(bitVector[index], bitIndex) == true) {
                incrementIndex();
            }
            return position();
        }
    }
}
