package cs307.first;

import threeohseven.BitVector;
import threeohseven.ImplementationError;

/**
 * This BitVector implementation is optimized to be space efficient for large, 
 * sparse bit vectors where the test and set functions are used most often. It stores
 * the vector as an array of indices of the 1-bits
 * 
 * @author -
 */
public class BitVectorB_46 implements BitVector {
	private boolean _checked;

	public int _min, _max; //min and max bit indices
	public int[] _bv; //array of indices of 1-bits
	public int _arraySize = 100; //increases if there are more than 100 1-bits
	public Builder builder;

	public BitVectorB_46() {
		super();
	}

	/**
	 * The constructor initializes all of the indices in the array to -1.
	 * 
	 * @param min vector index
	 * @param max vector index
	 * @param checked (true of bound checking is enforced)
	 */
	public BitVectorB_46(int min, int max, boolean checked) {
		_max = max; 
		_min = min;
		_bv = new int[_arraySize];
		_checked = checked;
		for (int i=0; i<100; i++) {
			_bv[i] = -1;
		}
	}

	/**
	 * Makes and returns a string representing the bit vector.
	 * @return string
	 */
	@Override
	public String toString() {
		char[] ch = new char[_max-_min+1];
		for (int i=0; i<_max-_min+1; i++) {
			ch[i] = '0';
		}
		for (int i=0; i<_arraySize; i++) {
			if (_bv[i] > _max || (_bv[i] < _min && _bv[i] != -1)) {
				throw new ImplementationError("out of bounds");
			}
			else if (_bv[i] != -1) {	
				ch[_bv[i]] = '1';
			}
		}
		String str = new String(ch);
		return str;
	}

	/**
	 * Given a BitVector bs, determines whether bs is compatible with 
	 * this. If so, it performs an AND operation on each pair of bits
	 * and stores the result in this BitVector.
	 * @param BitVector bs
	 */
	@Override
	public void and(BitVector bs) {
		BitVectorB_46 bvb = (BitVectorB_46) bs;
		if (this.builder.compatible(bvb.builder())) {
			for (int i=_min; i<=_max; i++) {
				if (this.test(i) && bvb.test(i)) {
					this.set(i);
				}
				else {
					this.unset(i);
				}
			}
		}
		else {
			throw new ImplementationError("incompatible BitVectors");
		}
	}

	/**
	 * Returns a copy of this BitVector's builder.
	 * @return BuilderS_mkorb
	 */
	@Override
	public Builder builder() {
		return builder;
	}

	/**
	 * Returns a copy of this BitVector by using the builder to create a new one.
	 * @return BitVectorS_mkorb
	 */
	@Override
	public BitVector copy() {
		return builder.buildBitVector(this);
	}

	/**
	 * Returns an Iterator that can be used to access this BitVector
	 * @return Iterator
	 */
	@Override
	public Iterator iterator() {
		return new IteratorB_46(this);
	}

	/**
	 * Negates the bit located at index and returns the old value. If there are
	 * _arraySize 1-bits and one more is needed, this function doubles the size
	 * of the index array and stores the new index at the first open position.
	 * 
	 * @return boolean
	 */
	@Override
	public boolean negate(int index) {
		if (index > _max || index < _min) {
			if (_checked) {
				throw new ImplementationError("Out of bounds");
			}
			else {
				return false;
			}
		}
		if (test(index)) {
			for (int i=0; i<_arraySize; i++) {
				if (_bv[i] == index) {
					_bv[i] = -1;
					return true;
				}
			}
		}
		else {
			int i=-1;
			while (_bv[++i] != -1) {
				if (i == _arraySize-1) {
					_arraySize *= 2;
					int[] bvdoubled = new int[_arraySize];
					for (int j=0; j<_arraySize/2; j++) {
						bvdoubled[j] = _bv[j];
					}
					for (int j=_arraySize/2; j<_arraySize; j++) {
						bvdoubled[j] = -1;
					}
					_bv = bvdoubled;
					i = _arraySize/2;
					break;
				}
			}
			_bv[i] = index;
			return false;
		}
		return false;
	}

	/**
	 * Given a BitVector bs, determines whether bs is compatible with 
	 * this. If so, it performs an OR operation on each pair of bits
	 * and stores the result in this BitVector.
	 * @param BitVector bs
	 */
	@Override
	public void or(BitVector bs) {
		BitVectorB_46 bvb = (BitVectorB_46) bs;
		if (this.builder.compatible(bvb.builder())) {
			for (int i=_min; i<=_max; i++) {
				if (this.test(i) || bvb.test(i)) {
					this.set(i);
				}
				else {
					this.unset(i);
				}
			}
		}
		else {
			throw new ImplementationError("incompatible BitVectors");
		}
	}

	/**
	 * Sets the bit at index and returns the old value.
	 * @return boolean
	 */
	@Override
	public boolean set(int index) {
		if (index > _max || index < _min) {
			if (_checked) {
				throw new ImplementationError("out of bounds");
			}
			else {
				return false;
			}
		}
		if (test(index)) {
			return true;
		}
		else {
			negate(index);
			return false;
		}
	}

	/**
	 * Returns the value of the bit at index.
	 * @return boolean
	 */
	@Override
	public boolean test(int index) {
		if (index > _max || index < _min) {
			if (_checked) {
				throw new ImplementationError("out of bounds");
			}
			else {
				return false;
			}
		}
		for (int i=0; i<_arraySize; i++) {
			if (_bv[i] == index) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Unsets the bit at index and returns the old value.
	 * @return boolean
	 */
	@Override
	public boolean unset(int index) {
		if (index > _max || index < _min) {
			if (_checked) {
				throw new ImplementationError("out of bounds");
			}
			else {
				return false;
			}
		}
		if (test(index)) {
			negate(index);
			return true;
		}
		else {
			return false;
		}
	}

	/**
	 * Given a BitVector bs, determines whether bs is compatible with 
	 * this. If so, it performs an XOR operation on each pair of bits
	 * and stores the result in this BitVector.
	 * @param BitVector bs
	 */
	@Override
	public void xor(BitVector bs) {
		BitVectorB_46 bvb = (BitVectorB_46) bs;
		if (this.builder.compatible(bvb.builder())) {
			for (int i=_min; i<=_max; i++) {
				if (this.test(i) ^ bvb.test(i)) {
					this.set(i);
				}
				else {
					this.unset(i);
				}
			}
		}
		else {
			throw new ImplementationError("incompatible BitVectors");
		}
	}
}
