/**
 * The BitVector interface for 307-PA1
 **/
package cs307.threeohseven;

/**
 * A BitVector maps values from an active range to booleans. The bit
 * vector's active range is denoted [low, high] such that low and high are
 * int value and low <= high. Any index within that range is valid.<p>
 *
 * The interface defines methods testing and setting individual positions
 * in the set's active range (<code>set, test, unset, negate</code>), as
 * well as some binary logical operations (<code>or, xor, and,
 * etc.</code>).<p>
 *
 * Each vector instance has a unique <code>Builder</code>, the builder is
 * the object that created this vector. Builders are immutable values thus
 * two vectors created by the same builder are guaranteed to be compatible.
 * More precisely, we say that two <code>BitVector</code> objects are
 * <em>compatible</em> if their builders are
 * <code>Buidler.compatible()</code>.<p>
 * 
 * Binary operations on <code>BitVector</code> are valid if and only if the
 * two BitVectors are compatible. Thus it is not valid to attempt to
 * <code>or</code> bit vector with different active ranges.
 *
 * With regards to error checking, an implementation may either be
 * <em>strict</em> or <em>relaxed</em>. A strict implementation must check
 * all operations and throw exception when an invalid operation is
 * attempted. A relaxed implementation may let some errors go
 * unreported. The motivation for the relaxed mode is to allow for some
 * common implementation idioms (for speed).  The relaxed mode should not
 * throw any undeclared exceptions.
 *
 * @author -
 * @author -
 * @version 1.0.0
 **/
public interface BitVector {
    /**
     * Test the value at <code>index</code>.  The operation is only valid
     * if <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return value at <code>index</code> 
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean test(int index);
    /**
     * Set the bit at <code>index</code>.  The operation is only valid if
     * <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean set(int index);
    /**
     * Unset the bit at <code>index</code>.  The operation is only valid if
     * <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean unset(int index);
    /**
     * Negate the bit at <code>index</code>.  The operation is only valid
     * if <code>index</code> is in the active range of the
     * <code>BitVector</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public boolean negate(int index);
    /**
     * Logical <code>and</code> of the two vectors. Receiver is updated in
     * place.  The operation is only valid if <code>index</code> is in the
     * active range of the <code>BitVector</code>.
     * @param bs the other <code>BitVector</code>.
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public void and(BitVector bs);
    /**
     * Logical <code>or</code> of the two vectors. Receiver is updated in
     * place.  The operation is only valid if the two
     * <code>BitVector</code> objects are compatible.
     * @param bs the other <code>BitVector</code>.
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public void or(BitVector bs);
    /**
     * Logical <code>xor</code> of the two vectors. Receiver is updated in
     * place.  The operation is only valid if the two
     * <code>BitVector</code> objects are compatible.
     * @param bs the other <code>BitVector</code>.
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code> is strict.
     **/
    public void xor(BitVector bs);
    /**
     * Returns a logically disjoint copy of the object.  The operation is
     * only valid if the two <code>BitVector</code> objects are compatible.
     * @param bs the other <code>BitVector</code>.
     * @throws ImplementationError if the operation is invalid and the
     * <code>BitVector</code>
     **/
    public BitVector copy();
    /**
     * Returns an iterator over this <code>BitVector</code>.
     **/
    public Iterator iterator();
    /**
     * Returns the builder that created this <code>BitVector</code>.
     **/
    public Builder builder();

    /**
     * Return a representation of this <code>BitVector</code> as a
     * string. The format is implementation specific.
     **/
    public String toString();

    //--------------------------------------------------------------

    /**
     * Builder creates new  <code>BitVector</code> objects. <p>
     *
     * @author -
     * @author -
     **/
    public interface Builder extends Director.Builder {
	/**
	 * Return a new empty BitVector.
	 **/
	public BitVector buildBitVector();
	/**
	 * Returns a new BitVector identical to <code>val</code>. This is a
	 * deep copy in the sense that there is no logical sharing of
	 * mutable state between the two objects. 
	 * @throws ImplementationError if the argument's builder
	 *  is not compatible  with the reciever.
	 **/
	public BitVector buildBitVector(BitVector val);
	/**
	 * Return true if the builders have the same internal state.
	 **/
	public boolean equals(Object other);
	/**
	 * Returns true if the two builders are compatible. This method is
	 * required to return true if the builders are
	 * <code>.equals</code>. For builders of different classes the
	 * method returns true when the implementation is able to convert between
	 * representations. <p>
	 **/
	public boolean compatible(Builder other);
    } // end of BitVector.Builder


    //--------------------------------------------------------------

    /**
     * Descriptor for a BitVector use case. This class provides an
     * implementation of Descriptors suitable for bit vectors.
     *
     * @author -
      **/
    public static final class Descriptor  
	implements Director.Descriptor {

	private final boolean optSpace_;
	private final boolean optSpeed_;
	private final int low_;
	private final int high_;
	private final boolean checked_;
	/**
	 * Make a descriptor for bit vectors from <code>low</code> to
	 * <code>high</code>.
	 * @param low lower bound on active range
	 * @param high higher bound on active range
	 **/
	public Descriptor(int low, int high) {
	    optSpace_ = false;
	    optSpeed_ = false;
	    low_ = low;
	    high_ = high;
	    checked_ = true;
	}

	/**
	 * Make a descriptor for bit vectors from <code>low</code> to
	 * <code>high</code>.
	 * @param low lower bound on active range
	 * @param high higher bound on active range
	 * @param optSpace optimize for space
	 * @param optSpeed optimize for speed
	 * @param checked  full bound checks required
	 **/
	public Descriptor(int low, 
			  int high,
			  boolean optSpace,
			  boolean optSpeed,
			  boolean checked) {
	    optSpace_ = optSpace;
	    optSpeed_ = optSpeed;
	    low_ = low;
	    high_ = high;
	    checked_ = checked;
	}

	/**
	 * Should storage space be minimized?
	 **/
	public boolean optimizeSpace() {
	    return optSpace_;
	}
	/**
	 * Should speed be maximized?
	 **/
	public boolean optimizeSpeed() {
	    return optSpeed_;
	}
	/**
	 * Lower bound on active range.
	 **/
	public int getLowerBound() {
	    return low_;
	}
	/**
	 * Upper bound on active range.
	 **/
	public int getUpperBound() {
	    return high_;
	}

	/**
	 * Returns true if bounds checks are required for all accesses. If
	 * this is set to false an implementation may choose not check some
	 * operations. Typically this allows removal of some redundant
	 * checks at the risk of overflowing. If the method returns true,
	 * the corresponding vectors are strict, otherwise they called
	 * relaxed.
	 **/
	public boolean boundCheckRequired() {
	        return checked_;
	}

	/**
	 * Return a short description of the requirements for BitVectors
	 * created by this descriptor.
	 **/
	public String getDescription() {
	    return "BitVector [" +low_+","+high_+"], opt space ="+optSpace_
		+" opt speed ="+optSpeed_;
	}
	/**
	 * Return true if the two descriptor are identical.
	 **/
	public boolean equals(Object o) {
	    if (o == this) return true;
	    if (o instanceof Descriptor == false) return false;
	    Descriptor descr = (Descriptor) o;
	    return 
		high_ == descr.high_ &&
		low_ == descr.low_ &&
		checked_ == descr.checked_ &&
		optSpeed_ == descr.optSpeed_ &&
		optSpace_ == descr.optSpace_;
	}

    } // end of BitVector.Descriptor


    //------------------------------------------------------------------------

    /**
     * Iterator over a BitVector. <p> An Iterator returns the values of
     * every index in the bit vector's active range in increasing
     * order. The iterator maintains a cursor into the vector, the cursor
     * is guaranteed to be larger or equal to the vector's lower bound.
     *
     * When created an interator points to first element in the active
     * range, so that <code>getNext()</code> returns the first value in the
     * mapping, and <code>position()</code> returns the lower bound of the
     * active range.
     *
     * @author -
     * @author -
     **/
    public interface Iterator {
	/**
	 * Are there more values in the active range?
	 * @return false if the end of the range has been reached.
	 **/
	public boolean hasNext();

	/**
	 * Returns the value for the current cursor, and advances the
	 * cursor by one. When the cursor overflows the active range
	 * (i.e. <code>hasNext()</code> is false) arbitrary values may be
	 * returned. Implementations are not required to throw an exception.
  	 * @return the value of the next index in range.
  	 * @throws NoSuchElementError if hasNext() == false and vector is strict.
	 **/
	public boolean getNext();

	/**
	 * The value of the cursor. This value may be larger than the
	 * vector's upper bound. In which case <code>hasNext()</code> will
	 * return false.
	 * @return current position.
	 **/
	public int position();

	/**
	 * Skips to the next set bit such that <code>getNext()</code> will
	 * return true, if none is found the cursor is moved to the end of
	 * the vector and <code>hasNext()</code> will return false from
	 * thereon.
	 * @return position of the first set bit, if there is none an
	 * invalid value will be returned (ie a value larger than the upper
	 * bound of the array).
	 **/
	public int skipToSet();

	/**
	 * Skips to the next unset bit such that <code>getNext()</code>
	 * will return false, if none is found the cursor is moved to the
	 * end of the vector and <code>hasNext()</code> will return false
	 * from thereon.
	 * @return position of the first unset bit, if there is none an
	 * invalid value will be returned (ie a value larger than the upper
	 * bound of the array).
	 **/
	public int skipToUnset();

    } // end of BitVector.Iterator

    public static class NoSuchElementError
	extends Error {
	public NoSuchElementError(String msg) {
	    super(msg);
	}
    } // end of NoSuchElementError

} // end of BitVector




