package cs307.first;

import threeohseven.BitVector;
import threeohseven.ImplementationError;

/**
 * A specific implementation of BitVector that has been optimized for space.
 * Instead of using booleans to store values, it packs 8 bits into a character.  This
 * method of storing data is a more space efficient but requires a little overhead in
 * terms of speed of reading and writing.
 */
public class BitVectorSpace_7 implements BitVector {
	private BitVector.Builder b;
	private int low, high;
	private boolean check;
	private char[] data;
	
	public BitVectorSpace_7 (BitVector.Builder b, int low, int high, boolean check) {
		this.b = b;
		this.low = low;
		this.high = high;
		this.check = check;
		
		int elements = (high-low+1) / 8;
		if ((high-low+1) % 8 > 0)
			elements++;
		
		data = new char [elements];
	}
	
	public boolean test(int index) {
		if (check) {
			if (index > high || index < low)
				throw new ImplementationError ("Out of bounds");
		}

		int sb = (index-low) / 8;
		int off = (index-low) % 8;
		
		return (data[sb] & (1 << off)) != 0;
	}
	
	public boolean set(int index) {
		if (check) {
			if (index > high || index < low)
				throw new ImplementationError ("Out of bounds");
		}

		int sb = (index-low) / 8;
		int off = (index-low) % 8;

		boolean old_value = (data[sb] & (1 << off)) != 0;
		
		data[sb] = (char) (data[sb] | (0x01 << off));
		
		return old_value;
	}
	
	public boolean unset(int index) {
		if (check) {
			if (index > high || index < low)
				throw new ImplementationError ("Out of bounds");
		}

		int sb = (index-low) / 8;
		int off = (index-low) % 8;

		boolean old_value = (data[sb] & (1 << off)) != 0;
		
		data[sb] = (char) (data[sb] & ~(1 << off));
		
		return old_value;
	}
	
	public boolean negate(int index) {
		if (check) {
			if (index > high || index < low)
				throw new ImplementationError ("Out of bounds");
		}

		int sb = (index-low) / 8;
		int off = (index-low) % 8;

		boolean old_value = (data[sb] & (1 << off)) != 0;

		if (old_value) {
			data[sb] = (char) (data[sb] & ~(1 << off));
		} else {
			data[sb] = (char) (data[sb] | (1 << off));
		}
		return old_value;
	}
	
	public void and(BitVector bs) {
		if (check) {
			if (!b.compatible (bs.builder ()))
				throw new ImplementationError ("Non-compatible objects");
		}
		
		for (int i = low; i <= high; i++) {
			if (test (i) && bs.test (i))
				set (i);
			else
				unset (i);
		}
	}
	
	public void or(BitVector bs) {
		if (check) {
			if (!b.compatible (bs.builder ()))
				throw new ImplementationError ("Non-compatible objects");
		}
		
		for (int i = low; i <= high; i++) {
			if (test (i) || bs.test (i))
				set (i);
			else
				unset (i);
		}
	}
	
	public void xor(BitVector bs) {
		if (check) {
			if (!b.compatible (bs.builder ()))
				throw new ImplementationError ("Non-compatible objects");
		}
		
		for (int i = low; i <= high; i++) {
			if (test (i) ^ bs.test (i))
				set (i);
			else
				unset (i);
		}
	}
	
	public BitVector copy() {
		BitVectorSpace_7 bvs = (BitVectorSpace_7) b.buildBitVector ();
		for (int i = 0; i < data.length; i++)
			bvs.data[i] = data[i];
		return (BitVector) bvs;
	}
	
	public Iterator iterator() {
		return new Iterator (this);
	}
	
	public Builder builder() {
		return b;
	}
	
	public String toString() {
		String ret = "";
		for (int i = high; i >= low; i--) {
			if (test (i))
				ret = ret + "1";
			else
				ret = ret + "0";
		}
		return ret;
	}
	
	public static class Iterator implements BitVector.Iterator {
		private int pos;
		private BitVectorSpace_7 bv;
		
		public Iterator (BitVectorSpace_7 bv) {
			this.bv = bv;
			pos = bv.low;
		}
		
		public boolean getNext () {
			return bv.test(pos++);
		}
		
		public boolean hasNext () {
			return (pos <= bv.high);
		}
		
		public int position () {
			return pos;
		}
		
		public int skipToSet () {
			while (pos <= bv.high) {
				if (bv.test (pos)) {
					return pos;
				}
				pos++;
			}
			return pos;
		}
		
		public int skipToUnset () {
			while (pos <= bv.high) {
				if (!bv.test (pos)) {
					return pos;
				}
				pos++;
			}
			return pos;
		}
	}

}
